//Doubly Linked list
//Just like a singly-linked list, but points to previous too
//Linked List

//Values stored in *non*-contiguous memory
// - grow and change as needed
//basically a bunch of instances and pointers

//Downsides
// - traverse list to get to data
// - retrieval always O(n)

//"Make me a doubly-linked list"
class DoublyLinkedList{
  constructor(){
    this.head = null;
  }
  traverse(fn){
    let current = this.head;
    let last = null;
    while(current){
      const result = fn(current, last);
      last = current;
      current = current.next;
      if(result) break;
    }
    return last;
  }
  insertAt(node, val){
    //1 -><- 2 -><- 4 -><- 3
    const newNode = {
      value: val,
      next: node.next,
      previous: node
    }
    if(node.next) node.next.previous = newNode;
    node.next = newNode;
    return newNode;
  }
  append(val){
    const newNode = {
      value: val,
      next: null,
      previous: null
    }
    if(!this.head){
      this.head = newNode;
    }else{
      //find the end of the list
      const lastNode = this.traverse(n => false);
      lastNode.next = newNode;
      newNode.previous = lastNode;
    }
    return newNode;
  }
  //A VERY POPULAR QUESTION
  remove(node){
    //1 <-> 2 <-> 3 <-> 4
    if(this.head.value === node.value){
      this.head = this.head.next;
    }else{
      if(node.previous){
        node.previous.next = node.next;
      }
      if(node.next){
        node.next.previous = node.previous;
      }
    }
  }
}

module.exports = DoublyLinkedList;
const list = new DoublyLinkedList();
const print = () => list.traverse(n => console.log(n.value));
list.append(1);
const two = list.append(2);
list.append(3);
print();
console.log("Inserting 5...");
list.insertAt(two, 5);
print();
console.log("Removing 2...");
list.remove(two);
print();
